#
# Copyright 2011 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Connect to the launchpad/gc dbus service and get a functional object.
"""

import os
import sys

from notmuch import Database
from MuchMuch.base import dbus_object, ROOT_ADDRESS

from dbus.exceptions import UnknownMethodException, DBusException

def dbus_cache(cls):
    instances = {}
    def getinstance(oid=None):
        address = oid
        if cls.address:
            address = cls.address
        elif oid and oid[0] != '/':
            # FIXED Item, tried to use cls.id_parent but the metaclass
            # Would always fail for some reason.
            root = MailRoom().obj
            if not cls.id_method:
                raise NotImplementedError("id_method property")
            address = str(getattr(root, cls.id_method)(oid))
        if not address:
            raise KeyError("DBus ID or Address not found! %s" % oid)
        if not instances.has_key(address):
            instances[address] = cls(address)
        return instances[address]
    return getinstance


class DBusClientObject(object):
    """Internal base class for dbus client objects."""
    id_method = None
    address = None

    def __init__(self, address=None):
        self.address = address or self.address
        self.obj = dbus_object(self.address)
        self.oid = self.obj.GetId()


class MailRoom(DBusClientObject):
    """Operate all mail services from here"""
    address = ROOT_ADDRESS

    def mail_slots(self):
        """Return a list of MailSlot objects"""
        for address in self.obj.GetMailSlots():
            yield MailSlot(address)

    def mail_accounts(self):
        """Return a list of MailAccount objects"""
        for address in self.obj.GetAccounts():
            yield MailAccount(address)


@dbus_cache
class MailAccount(DBusClientObject):
    """Open a mail account to control the account access"""
    id_method = 'GetAccount'


@dbus_cache
class MailSlot(DBusClientObject):
    """Open a mailslot for read and write access to emails."""
    id_method = 'GetMailSlot'

    def __init__(self, address):
        self.address = address or self.address
        self.obj = dbus_object(self.address)
        # Gather some data for id should the write db die
        self.path = self.obj.GetPath()
        self.slot_id = self.obj.GetId()

    def search(self, query, sort='newest'):
        """
        Perform a search using the provided query string.
        
        query   - Query (see xapian for query syntax)
        sort    - Newest, Oldest, Msgid or Unsort

        Returns a query id for use with get_messages and get_threads.
        """
        return self.obj.Search(query, sort)

    def get_messages(self, qid, count=20, out=None):
        """
        Get the resulting messages from a search.
        
        qid     - Query Id from previous call to search.
        count   - Number of messages to fetch, 0 is all.
        out     - Type of output:
            * None  - will return (msgid, filename) tuples
            * True  - will return list of message ids
            * False - will return list of filenames

        Return a list of messages found
        """
        ret = self.obj.GetMessages(qid, count)
        if out != None:
            result = []
            for (msgid, fn) in ret:
                result.append(out and msgid or fn)
            ret = result
        return ret

    def messages(self, query, sort='newest', out=None):
        """
        Combination method of search + get_messages,

        Returns all available messages for query.
        """
        return self.get_messages(self.search(query, sort), count=0, out=out)


    def deliver(self, filename, new=False, tags=None):
        """
        Deliver a message to this mailslot:

        filename - filename inside or outside the mailslot path.
        new      - Is this email new (freshly delivered?).
        tags     - Tags to send into the datase on delivery.
        """
        if filename[0] != '/':
            expath = os.path.join(self.path, filename)
            if os.path.exists(expath):
                filename = expath
        if not tags:
            tags = new and ['new'] or ['read']
        elif new and 'new' not in tags:
            tags.append('new')
        msgid = self.obj.Deliver(filename, tags)
        # Return the unique message id
        return msgid

    def delete_message(self, msgid, purge=True):
        """
        Delete a message from the database (by filename or message id).

        purge the original filename if needed (default: True).
        """
        return self.obj.DeleteMessage(msgid, purge)

    def full_scan(self, path=None, tags=None):
        """Scan the given path for new emails"""
        path = path or self.path
        tags = tags or ['read']
        return self.obj.FullScan(path, tags)

    def add_tag(self, msgid, tag):
        """Add a single tag to the email"""
        return self.obj.AddTag(msgid, tag)

    def add_tags(self, msgid, tags):
        """Add multiple tags to an email"""
        return self.obj.AddTags(msgid, tags)

    def remove_tag(self, msgid, tag):
        """Remove a given tag from the email"""
        return self.obj.RemoveTag(msgid, tag)

    def remove_tags(self, msgid):
        """Remove all tags from a single email"""
        return self.obj.RemoveTags(msgid)

    def delete_tag(self, tag):
        """Remove all instances of a tag from the slot"""
        return self.obj.DeleteTag(tag)

    def connect_to_signal(self, signal, method):
        """Connect to one of the mail slot signals."""
        return self.obj.connect_to_signal(signal, method)

    def connect(self, *args):
        """Same as connect_to_signal"""
        return self.connect_to_signal(*args)

