#
# Copyright 2011 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Control email accounts, deleting accounts, making new accounts
and even finding out what type of accounts are posible.

"""
import os
import sys
import logging
import subprocess

from MuchMuch import __testing__
from MuchMuch.config import ConfigFile
from MuchMuch.base import DBusObject, SOA, dbus_method, dbus_signal

class AccountLoadError(IOError):
    def __init__(self, msg, config):
        msg = msg + ", Ignoring account: " + str(config)
        super(type(self), self).__init__(msg)


class MailAccount(DBusObject):
    """Access to email accounts via DBus base class."""
    def __init__(self, types, filename):
        self.retrys = 0
        self.filename = filename
        try:
            self.config = ConfigFile(filename)
        except ValueError:
            raise AccountLoadError("Account conf invalid", filename)
        
        if not self.config or not self.config.has_key('account') or \
            not self.config['account'].has_key('id'):
            raise AccountLoadError("Invalid configuration", filename)

        self.type_id = self.config['account'].get('type', 'Missing')
        if not types.has_key(self.type_id):
            raise AccountLoadError("No plugin '%s'" % (self.type_id), filename)
        self._type = types[self.type_id]
        self._cmd = self._type['program']
        self.oid = self.config['account']['id']
        DBusObject.__init__(self, 'Account/' + self.oid)
        self.load_process()

    @dbus_method(SOA, in_signature='', out_signature='s')
    def GetId(self):
        """Returns the Id of the account"""
        return self.oid

    def load_process(self):
        """Load the process and track it"""
        self.watcher = ThreadWatcher([self._cmd, self.filename], self._died)
        self.watcher.start()

    def _died(self):
        """Called when the process died."""
        if self.retrys > 5:
            time.sleep(0.5)
            self.retrys += 1
            logging.warn("Retrying Plugin: %s" % self.oid)
            self.load_process()
        else:
            logging.warn("Plugin Died: %s" % self.oid)



import threading
import signal
import atexit

class ThreadWatcher(threading.Thread):
    """Locks on a process and callbacks on return"""
    def __init__(self, cmd, callback, args=None):
        self.killed = False
        self.proc = None
        self.cmd = cmd
        self.callback = callback
        self.args = args or []
        self.logging = None
        if __testing__:
            end = os.path.basename(cmd[1]).split('.')[0]
            self.logging = open('plugin-%s.log' % (end), 'w')
        threading.Thread.__init__(self)

    def run(self):
        atexit.register(self.kill)
        self.proc = subprocess.Popen(self.cmd,
            stderr=self.logging,
            stdout=self.logging)
        self.proc.wait()
        if not self.killed:
            self.callback(*self.args)

    def kill(self):
        self.killed = True
        try:
            os.kill(self.proc.pid, signal.SIGTERM)
        except OSError:
            pass

