#
# Copyright 2011 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Provide Internal management functionality for the email accounts services."""
import os
import sys
import logging
import dircache
import json

from subprocess import Popen, PIPE

from MuchMuch import PLUGIN_DIR
from MuchMuch.mailaccount import MailAccount
from MuchMuch.config import ConfigFile, CONF_DIR

class AccountManager(object):
    """Controller for Email Accounts"""
    _types = None

    def __init__(self):
        self._list = []
        self._lookup = {}
        self.load_all()

    def get_account(self, aid):
        """Return the account given an id"""
        if self._lookup.has_key(aid):
            return self._lookup[aid]
        return None

    def create_account(self, type_id, email):
        """Create a new email account of the type specified"""
        if not self._types.has_key(type_id):
            raise KeyError("Please use a valid type_id, not: %s" % type_id)
        account_id = email.replace('@', '_').replace('.', '_')
        config_file = os.path.join(CONF_DIR, account_id+'.conf')

        # Generate a new configuration file
        config = ConfigFile(config_file)
        config['account'] = {
            'type': type_id,
            'id': account_id,
        }
        config.save()
        del config

        # Load
        self.load_item(config_file)
        return account_id

    @property
    def types(self):
        """Return a list of account types, full info."""
        if not self._types:
            self._types = {}
            # Look in the account plugins directory for available accounts.
            for program in dircache.listdir(PLUGIN_DIR):
                info = self.type_info(program)
                if not info:
                    logging.warn("Ignoring email account program: %s" % program)
                    continue
                self._types[info['id']] = info
                logging.warn("Found account type: %s (%s)" % (
                    info['name'], info['version']))
        return self._types

    def type_info(self, program):
        """Return the information accociated with a program"""
        path = os.path.join(PLUGIN_DIR, program)
        if not os.access(path, os.X_OK):
            logging.warn("Not executable program: %s" % program)
            return None
        try:
            # run the program without any args to return data.
            output = Popen([path], stdout=PIPE).communicate()[0]
            info = json.loads(output)
            info['program'] = path
        except TypeError:
            logging.warn("Errors in returned data." % program)
            return None
        # Do some checking on the args to make sure it's valid.
        if not info.has_key('name') or not info.has_key('id') or \
          not info.has_key('version'):
            logging.warn("Returned data isn't complete." % program)
            return None
        return info

    def load_all(self):
        """go through all saved configurations and load their account processes"""
        if not os.path.exists(CONF_DIR):
            os.makedirs(CONF_DIR)
        for item in dircache.listdir(CONF_DIR):
            if item[-5:] != '.conf':
                continue
            config_file = os.path.join(CONF_DIR, item)
            self.load_item(config_file)
    
    def load_item(self, config_file):
        try:
            account = MailAccount(self.types, config_file)
        except Exception, error:
            logging.warn("Error loading MailAccount: %s" % error)
            raise
        else:
            self._list.append(account)
            self._lookup[account.oid] = account
        return True

