#
# Copyright 2011 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Provide dbus objects for running the core daemon email service object."""

import os
import sys
import logging

from notmuch import globals
from MuchMuch.base import DBusObject, SOA, dbus_method, dbus_signal
from MuchMuch.mailslots import MailSlotManager
from MuchMuch.mailaccounts import AccountManager

class EmailService(DBusObject):
    """Central control point for all email related elements."""
    def __init__(self):
        self.slots = MailSlotManager()
        DBusObject.__init__(self, None)
        self.accounts = AccountManager()

    @dbus_method(SOA, in_signature='', out_signature='s')
    def GetId(self):
        """Returns a static id 'root' for this manager service object."""
        return 'root'

# ======= Account Management ========= #

    @dbus_method(SOA, in_signature='', out_signature='ao')
    def GetAccounts(self):
        """Return a list of email accounts"""
        result = []
        for obj in self.accounts._list:
            result.append(obj.address)
        return result

    @dbus_method(SOA, in_signature='s', out_signature='o')
    def GetAccount(self, account_id):
        """Return the email account object by it's Id"""
        obj = self.accounts.get_account(account_id)
        if not obj:
            raise KeyError("Can't find email account: %s" % account_id)
        return obj.address

    @dbus_method(SOA, in_signature='ss', out_signature='b')
    def CreateAccount(self, type_id, email):
        """Create a new email account"""
        acc_id = self.accounts.create_account(type_id, email)
        if acc_id:
            self.AccountCreated(acc_id)
            return True
        return False


    @dbus_method(SOA, in_signature='s', out_signature='b')
    def DeleteMailSlot(self, account_id):
        """Delete an Email Account"""
        return False

    @dbus_method(SOA, in_signature='', out_signature='a(ss)')
    def AccountTypes(self):
        """Return a list of account types, (id, name) pairs."""
        result = []
        for info in self.accounts.types:
            result.append(info['id'], info['name'])
        return result

# ======= MailSlot Management ========= #

    @dbus_method(SOA, in_signature='', out_signature='ao')
    def GetMailSlots(self):
        """Return a list of mailslots (dbus method)"""
        result = []
        for obj in self.slots.slots:
            result.append(obj.address)
        return result

    @dbus_method(SOA, in_signature='s', out_signature='o')
    def GetMailSlot(self, pathorid):
        """Return a mailslot given either it's path or slot_id"""
        obj = self.slots.get_mail_slot(pathorid)
        if not obj:
            raise KeyError("Can't find MailSlot: %s" % pathorid)
        return obj.address

    @dbus_method(SOA, in_signature='s', out_signature='b')
    def CreateMailSlot(self, path):
        """Create a new mailslot (dbus method)"""
        slot_id = self.slots.create_mail_slot(path)
        if slot_id:
            self.MailSlotCreated(slot_id)
            return True
        return False

    @dbus_method(SOA, in_signature='sb', out_signature='b')
    def DeleteMailSlot(self, pathorid, purge=False):
        """Return a list of mailslots (dbus method)"""
        slot_id = self.slots.delete(pathorid, purge)
        if slot_id:
            self.MailSlotDeleted(slot_id)
            return True
        return False

# ======= Signals ========= #

    @dbus_signal(SOA, signature='s')
    def AccountCreated(self, account_id):
        """Signal to record account creation."""

    @dbus_signal(SOA, signature='s')
    def AccountDeleted(self, account_id):
        """Signal to record account deletion."""

    @dbus_signal(SOA, signature='s')
    def MailSlotCreated(self, slot_id):
        """Signal to record mail slot creation."""

    @dbus_signal(SOA, signature='s')
    def MailSlotDeleted(self, slot_id):
        """Signal to record mail slot deletion."""


