#
# Copyright 2011 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Provide access to indervidual email slots (databases) over dbus"""

import os
import sys
import logging
import random
import datetime
import gobject

from notmuch import Database, Query, Messages, Threads
from notmuch.message import STATUS
from notmuch.globals import NotmuchError
from MuchMuch.base import DBusObject, SOA, dbus_signal, dbus_method

QUERYSORT = {
    'oldest': Query.SORT.OLDEST_FIRST,
    'newest': Query.SORT.NEWEST_FIRST,
    'msgid' : Query.SORT.MESSAGE_ID,
    'unsort': Query.SORT.UNSORTED,
}
"""
Sorting variables:

In searches you can use:
 'oldest' - Sort Oldest First
 'newest' - Sort Newest First
 'msgid'  - Sort by Message Id
 'unsort' - No sorting (default)
"""

class DeliveryError(Exception):
    pass

class MessageNotFound(KeyError):
    pass

class UnknownSortMethod(KeyError):
    pass

class MailSlot(DBusObject):
    """Contact a mailslot directory with a unique id."""
    def __init__(self, slot_id, path):
        self.mid = slot_id
        self.path = unicode(os.path.abspath(path))
        self.xpath = os.path.join(self.path, '.notmuch')

        if not os.path.exists(self.xpath):
            self.db = Database(self.path, create=True)
        else:
            readwrite = Database.MODE.READ_WRITE
            self.db = Database(self.path, mode=readwrite)

        self.queries = { '': None }
        DBusObject.__init__(self, 'Slot/'+slot_id)

        self.timeout = 5 * 60 * 100 # 5 Minuets
        gobject.timeout_add(self.timeout, self.auto_flush)

    def auto_flush(self):
        """Internal method for saving changes to disk"""
        logging.warn("Flushing changes to disk.")
        self.flush()
        gobject.timeout_add(self.timeout, self.auto_flush)

    def flush(self):
        """Flush the database to disk"""
        if hasattr(self.db, 'flush'):
            self.db.flush()
        else:
            # The 0.8 notmuch doesn't have flush() yet.
            del self.db
            self.db = Database(self.path, mode=readwrite)

    def resolve_filename(self, filename):
        """Internal method used for delivery and other filename methods"""
        match1 = os.path.abspath(filename)
        if self.path in match1:
            return match1
        match2 = os.path.join(self.path, filename)
        if os.path.exists(match2):
            return match2
        if os.path.exists(match1):
            return match1
        return None

    def move_email(self, oldname):
        """Moves an email file to this mailslot directory"""
        if not os.path.isfile(oldname):
            logging.warn("MOVE: Couldn't move file '%s'" % oldname)
            return ''
        filepart = os.path.basename(oldname)
        # $path/$YY/$MM/$DD-$filename.eml
        day = datetime.datetime.today()
        path = os.path.join(self.path, "%02d" % day.year, "%02d" % day.month)
        if not os.path.isdir(path):
            os.makedirs(path)
        filename = "%02d-%s.eml" % (day.day, filepart.replace('.eml', ''))
        newname = os.path.join(path, filename)
        os.rename(oldname, newname)
        return newname

    @dbus_method(SOA, in_signature='', out_signature='s')
    def GetId(self):
        """Return this MailSlot's Unique Identifier"""
        return self.mid

    @dbus_method(SOA, in_signature='', out_signature='s')
    def GetPath(self):
        """Return this MailSlot's Directory Path"""
        return self.path

    @dbus_method(SOA, in_signature='s', out_signature='s')
    def FindMessage(self, msgid):
        """Return msgid for a filename and a filename for a msgid."""
        return self.find_message(msgid)[1]

    def find_message(self, msgid):
        """Internal method"""
        filename = self.resolve_filename(msgid)
        if filename:
            msg = self.db.find_message_by_filename(filename)
            if not msg:
                raise MessageNotFound("Using Filename: '%s'" % filename)
            # Return the message id when we know the filename
            return (msg, msg.get_message_id())
        else:
            msg = self.db.find_message(msgid)
            if not msg:
                raise MessageNotFound("Using ID: '%s'" % str(msgid))
            return (msg, msg.get_filename())

    def remove_message(self, filename, purge=False):
        """Internal method for removing and purging an email"""
        self.db.remove_message(filename)
        if purge:
            os.unlink(filename)
        return self.find_message(filename)[0]

    @dbus_method(SOA, in_signature='sas', out_signature='s')
    def Deliver(self, fn, tags):
        """Deliver a message."""
        filename = self.resolve_filename(fn)
        if not filename:
            raise DeliveryError("Can't find message: %s" % fn)

        # Move outside emails to inside the directory.
        if filename and not self.path in filename:
            filename = self.move_email(filename)

        logging.warn("Adding Message: %s" % filename)
        try:
            (msg, status) = self.db.add_message(filename)
        except NotmuchError, error:
            raise DeliveryError(str(error)+": %s" % filename)

        if status == 6 and 'new' in tags:
            raise DeliveryError(STATUS.status2str(status))
        logging.warn(STATUS.status2str(status))

        self.AddTags(msg, tags)
        # Format the message data for returning
        msgid = msg.get_message_id()
        if msg:
            self.MessageDelivered(msgid, filename)
            if 'new' in tags:
                self.NewMessage(msgid, filename)
        return msgid

    @dbus_method(SOA, in_signature='sb', out_signature='b')
    def DeleteMessage(self, idorfile, purge=False):
        """Remove a message from the database"""
        msg = None
        filename = self.resolve_filename(idorfile)
        msg = self.find_message(filename)[0]
        if not msg:
            raise Exception("Message was never in database: %s" % filename)

        if filename:
            # Removing a single filename
            self.remove_message(filename, purge)
        else:
            # Input was a message id, so remove all filenames
            for filename in msg.get_filenames():
                self.remove_message(filename, purge)
        # Send deleted signal over dbus
        self.MessageDeleted(msg.get_message_id())
        return True

    @dbus_method(SOA, in_signature='ss', out_signature='b')
    def AddTag(self, msgid, tag):
        """Add a tag to a an email"""
        msg = self.find_message(msgid)[0]
        if msg:
            msg.add_tag(tag)
            return True
        return False

    @dbus_method(SOA, in_signature='sas', out_signature='b')
    def AddTags(self, msgid, tags):
        """Add multiple tags to an email"""
        if isinstance(msgid, basestring):
            msg = self.find_message(msgid)[0]
        else:
            msg = msgid
        if msg:
            msg.freeze()
            for tag in tags:
                msg.add_tag(tag)
            msg.thaw()
            return True
        return False

    @dbus_method(SOA, in_signature='ss', out_signature='b')
    def RemoveTag(self, msgid, tag):
        """Remove a tag from an email"""
        msg = self.find_message(msgid)[0]
        if msg:
            msg.remove_tag(tag)
            return True
        return False

    @dbus_method(SOA, in_signature='s', out_signature='b')
    def RemoveTags(self, msgid):
        """Remove all emails from an email"""
        msg = self.find_message(msgid)[0]
        if msg:
            msg.remove_all_tags()
            return True
        return False

    @dbus_method(SOA, in_signature='s', out_signature='b')
    def DeleteTag(self, tag):
        """Delete every mention of a tag"""
        query  = self.db.create_query('tag:%s' % tag)
        for msg in query.search_messages():
            msg.remove_tag(tag)
        return True

    @dbus_method(SOA, in_signature='sas', out_signature='a(ss)')
    def FullScan(self, inpath, tags):
        """Scan the email slot directory for files not in the database."""
        if not os.path.exists(inpath):
            raise IOError("Scan directory not found: %s" % path)
        result = []
        for root, subFolders, files in os.walk(inpath):
            if '.notmuch' in root:
                continue
            for f in files:
                path = os.path.join(root, f)
                try:
                    result.append( (self.Deliver(path, tags), path) )
                except DeliveryError:
                    logging.warn("Scan: Skipping delivery of %s" % path)
                    pass
        return result

    @dbus_method(SOA, in_signature='ss', out_signature='s')
    def Search(self, query, sort):
        """Create a search query and return the query id for later use."""
        # Sort disabled for now
        qid = ''
        while self.queries.has_key(qid):
            qid = ''.join(random.sample('ABCDEF1234567890', 12))
        self.queries[qid] = self.db.create_query(query)

        if QUERYSORT.has_key(sort):
            sort = QUERYSORT[sort]
        if not sort in QUERYSORT.values():
            raise UnknownSortMethod("Can not sort with '%s'." % str(sort))

        self.queries[qid].set_sort(sort)
        return qid

    @dbus_method(SOA, in_signature='sd', out_signature='a(ss)')
    def GetMessages(self, qid, count=20):
        """Get messages from the search query using the generated id"""
        def proc(msg):
            return msg
        def gen(query):
            return query.search_messages()
        return self.process_search(qid, count, proc, gen, Messages)

    @dbus_method(SOA, in_signature='sd', out_signature='a(ss)')
    def GetThreads(self, qid, count=20):
        """Get threads only from the search query using the generated id"""
        def proc(thread):
            return thread.get_toplevel_messages()[0]
        def gen(query):
            return query.search_threads()
        return self.process_search(qid, count, proc, gen, Threads)
        return result

    def process_search(self, qid, count, proc, gen, cls):
        """Internal method for processing a search query"""
        if not self.queries.has_key(qid):
            raise KeyError("QueryID not found.")

        if isinstance(self.queries[qid], Query):
            self.queries[qid] = gen(self.queries[qid])
        if not isinstance(self.queries[qid], cls):
            raise KeyError("Can't get messages in this way, wrong type.")

        result = []
        for thing in self.queries[qid]:
            msg = proc(thing)
            result.append((msg.get_message_id(), msg.get_filename()))
            count -= 1
            # If the count was already 0 then the number is now -1
            # This means 0 is the sign for 'all available' i.e. no limit.
            if count == 0:
                break
        if count:
            del self.queries[qid]
        return result

    @dbus_method(SOA, in_signature='s', out_signature='d')
    def GetCount(self, qid):
        """return the estimatred number of messages for a search"""
        if not self.queries.has_key(qid):
            raise KeyError("QueryID not found.")
        query = self.queries[qid]
        return query.count_messages()

    @dbus_signal(SOA, signature='ss')
    def NewMessage(self, msgid, filename):
        """Signal which specifies a new (really new) message."""

    @dbus_signal(SOA, signature='ss')
    def MessageDelivered(self, msgid, filename):
        """Signal which specifies a message is added to the database."""

    @dbus_signal(SOA, signature='s')
    def MessageDeleted(self, msgid):
        """Signal when any email is deleted."""

    @dbus_signal(SOA, signature='s')
    def MessageChanged(self, msgid):
        """Signal when any email's status changes."""

