#
# Copyright 2011 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Provide an Internal Manager for MailSlots, used by the MailRoom"""

import os
import sys
import uuid
import shutil
import logging

from MuchMuch.config import ConfigFile
from MuchMuch.mailslot import MailSlot

class MailSlotManager(object):
    """MailSlot Manager singleton"""
    def __init__(self):
        self.config = ConfigFile()
        self.slots = []
        self.slotids = {}
        self.slotpaths = {}
        self.load_slots()

    def load_slots(self):
        """Loads from the configuration, all existing mailslots"""
        for (mid, path) in self.config.get('slots', []):
            self.load_slot(mid, path)

    def load_slot(self, mid, path):
        """Load any mailslot (create or existing"""
        obj = None
        path = os.path.abspath(path)
        if not os.path.exists(path):
            logging.error("Can't load MailSlot, path %s doesn't exist" % path)
        else:
            try:
                obj = MailSlot(mid, path)
            except globals.NotmuchError:
                logging.warn("Couldn't load database %s (%s)" % (mid, path))
        if obj and obj.GetId():
            self.slots.append(obj)
            self.slotids[mid] = obj
            self.slotpaths[path] = obj
            return obj
        else:
            logging.error("Can't load mailslot @ '%s'" % path)

    def create_mail_slot(self, path):
        if path[0] != '/':
            path = os.path.abspath(path)
        if self.slotpaths.has_key(path):
            logging.warn("Not creating mailslot (existing): %s" % path)
            return False
        if os.path.exists(os.path.join(path, '.notmuch')):
            logging.warn("Adding existing mailslot at %s" % path)
        else:
            logging.warn("Creating new mailslot at %s" % path)
        name = os.path.basename(path)
        mid = name + '_' + str(uuid.uuid1()).split('-')[0]
        obj = self.load_slot(mid, path)
        if obj != None:
            if not self.config.has_key('slots'):
                self.config['slots'] = []
            self.config['slots'].append([mid, path])
            self.config.save()
            return mid
        return None

    def get_mail_slot(self, pathorid):
        if self.slotids.has_key(pathorid):
            return self.slotids[pathorid]
        path = os.path.abspath(pathorid)
        if self.slotpaths.has_key(path):
            return self.slotpaths[path]

    def remove_config(self, obj):
        for x in range(len(self.config['slots'])):
            (t_mid, t_path) = self.config['slots'][x]
            if t_mid == obj.GetId():
                self.config['slots'].pop(x)
                self.config.save()
                return t_path
        return False

    def delete(self, pathorid, purge=False):
        """Delete a mail slot"""
        obj = self.get_mail_slot(pathorid)
        if obj:
            slot_id = obj.GetId()
            t_path = self.remove_config(obj)
            if purge and t_path:
                logging.warn("Removing tree: %s" % t_path)
                shutil.rmtree(t_path)
            self.slots.remove(obj)
            self.slotids.pop(slot_id)
            self.slotpaths.pop(obj.GetPath())
            del obj.db
            del obj
            return slot_id
        logging.warn("Can't delete '%s', no object." % pathorid)
        return None

