#
# Copyright (C) 2010 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Base test classes to poke at the dbus api.
"""

from base_test import Pipette, TestBox

import os
import gobject
import threading
import subprocess
import unittest
import logging
import atexit
import signal
import time

from dbus.exceptions import DBusException
from MuchMuch.base import ROOT_ADDRESS, dbus_object


class DBusClientService(threading.Thread):
    """
    Manages a process call which loads an external program in a threaded
    non-blocking way. As well as the load, it will exit the program (SIGKILL)
    at this program's exit signal. Perfect for dbus clients (used for testing)

    script    - Script / program to load
    dbus      - DBus object to wait for before sucessful load is considered.
    callback  - Method to run when program is finished, waits if no callback.
    timeout   - Automatically kill service after this amount of ms.
    log       - Save stderr to this filename.
    """
    def __init__(self, script, dbus, callback=None, log=None, **kwargs):
        super(type(self), self).__init__(**kwargs)
        self.script = script
        self.dbus = dbus
        self.callback = callback
        self.dbus = dbus
        self.log = log
        self.lock = False
        self.service = None
        self.obj = None

    def run(self):
        """Run the gobject loop and wait for signal"""
        atexit.register( self.kill )
        try:
            self.obj = dbus_object(self.dbus)
        except DBusException, error:
            self.start_service()
        if self.obj and self.callback:
            self.callback(self.dbus, self.obj)
        return self.obj

    def kill(self):
        if self.service:
            os.kill(self.service.pid, signal.SIGTERM)
            if isinstance(self.log, file):
                self.log.close()
            self.service = None

    def dbus_active(self):
        """Returns true if the service is still active"""
        if self.service:
            self.service.poll()
            if self.service.returncode == None:
                return True
        else:
            logging.warn("Can't find service!")
        return False

    def clean_script(self, script):
        """Return the abspath of the script"""
        if script[0] != '/':
            if script[0] != '.':
                script = os.path.join('.', script)
        return os.path.abspath(script)

    def start_service(self):
        """Start the dbus service for groundcontrol"""
        loop_itr = 0     # Iterate
        loop_dur = 0.1   # Every tenth of a second
        loop_res = None  # To get a result
        loop_max = 100   # Until 10 seconds
        if self.log:
            self.log = open(self.log, 'w')

        if isinstance(self.script, basestring):
            self.script = self.clean_script(self.script)
        else:
            self.script[0] = self.clean_script(self.script[0])

        # Launch a process to handle the dbus service
        self.service = subprocess.Popen(self.script, stderr=self.log)

        while loop_res == None:
            try:
                loop_res = dbus_object(self.dbus)
            except DBusException, error:
                # Continue with error if we've reached our limit or if it's different
                if 'ServiceUnknown' not in error.get_dbus_name() \
                  or loop_itr > loop_max:
                    raise
                # Go to sleep if dbus isn't ready
                loop_itr += 1
                time.sleep(0.1)
        self.obj = loop_res


class SignalThread(threading.Thread):
    """Simple gobject loop thread for catching signals"""
    def __init__(self):
        self.called = False
        self.result = {}
        self.loop = gobject.MainLoop()
        # Now continue with the threads initialisation
        threading.Thread.__init__(self)

    def run(self):
        """Run the gobject loop and wait for signal"""
        gobject.timeout_add(5000, self.kill)
        self.loop.run() # Blocking

    def all_stop(self):
        """Quit loop and all halt!"""
        self.stop()
        while self.is_running():
            print "MainLoop is still running...\n"
            time.sleep(1)

    def stop(self):
        """Quit the gobject mainloop"""
        self.called = True
        time.sleep(0.01)
        self.loop.quit()

    def kill(self):
        """Kill the gobject mainloop"""
        print "Forcing a signal mainloop to end!"
        self.stop()

    def callback(self, name):
        """Generic Signal Callback"""
        def internal(*args):
            self.called = True
            self.result[name] = args
        return internal

    def wait(self, name, timer=20):
        """Wait for any signal to complete"""
        while not self.called and timer > 1:
            time.sleep(0.1)
            timer -= 1
        return self.result.get(name, None)

    def reset(self):
        """Reset the callback variables"""
        self.called = False
        self.result = {}

    def is_running(self):
        """Return's true if thigns are still running"""
        return self.loop.is_running()

# We need to track when we have a dbus service error, so
# each test doesn't wait forever for the service to start.
ERROR = None

class DBusTestBase(TestBox):
    """Base test case for testing dbus methods."""
    def setUp(self):
        """
        Do NOT override this method, DBus and Threading!

        Use 'creation()' instead.
        """
        # Don't go any further if we have errors
        global ERROR
        self.assertFalse(ERROR, ERROR)
        self.service_root = find_service()
        self.assertFalse(ERROR, ERROR)
        self.assertTrue(self.service_root, "Previous dbus start failed.")

        self.mainloop = SignalThread()
        self.mainloop.start()
        if hasattr(self, 'creation'):
            try:
                self.creation()
            except:
                # We need to protect the mainloop as much as possible.
                self.mainloop.stop()
                raise

    def tearDown(self):
        """
        Do NOT override this method, DBus and Threading!

        Use 'destruction()' instead
        """
        if self.mainloop:
            self.mainloop.reset()
            self.mainloop.all_stop()
        if hasattr(self, 'destruction'):
            self.destruction()

    def failUnlessDBusRaises(self, err_name, cmd, *args, **kwargs):
        """Test a dbus call to make sure it throws the right exeception"""
        try:
            cmd(*args, **kwargs)
        except DBusException, error:
            name = error.get_dbus_name().replace('org.freedesktop.DBus.Python.', '')
            self.assertEqual(err_name, name,
                "Wrong DBus Error found '%s' != '%s': %s" %
                (name, err_name, error.get_dbus_message())
            )

    def signal_setup(self, obj, signal):
        """Bind a signal, expect caller to deal with waiting and collecting results"""
        obj.connect_to_signal(signal, self.mainloop.callback(signal))

    def signal_test(self, obj, signal, method):
        """Bind to a signal and wait for the returning data"""
        self.signal_setup(obj, signal)
        method_result = method()
        loop_result = self.signal_wait(signal)
        return (method_result, loop_result)

    def signal_wait(self, signal, timeout=20):
        """Wait for some data from a singal, caution, blocking."""
        if timeout == 0:
            return self.mainloop.result.get(signal)
        return self.mainloop.wait(signal, timeout)


def find_service():
    """Attempt to find an externally running daemon before starting
    a new one. Remember that test daemon has a different address."""
    global ERROR
    log_file = 'service.log'
    try:
        service = DBusClientService('execute_service.py', ROOT_ADDRESS, log=log_file)
        service.start()
        service.join(1000)
        if not service.isAlive():
            return service.obj
    except:
        ERROR = "DBus Error Detected!\n\n"
        if os.path.exists(log_file):
            with open(log_file, 'r') as fhl:
                ERROR += fhl.read()


