#
# Copyright 2011, Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Test listing and access to mailslots via dbus.
"""

import os
import shutil
import time
import dbus

from base_test import Pipette, prepare_data
from base_dbus_test import DBusTestBase, dbus_object
from MuchMuch import client
from MuchMuch.client import MailSlot
from MuchMuch.config import CONF_FILE

class CreationTestCase(DBusTestBase):
    """Testcase for creating mailslots via dbus"""
    def create_mailslot(self, path):
        """Creates a mailslot per test."""
        if not os.path.exists(path):
            os.makedirs(path)
        return self.service_root.CreateMailSlot(path)

    def test_01_create(self):
        """Create a mailslot"""
        path = 'data/mailslot1'
        success = self.create_mailslot(path)
        rpath = os.path.join(path, '.notmuch', 'xapian', 'record.DB')
        self.assertTrue(success)
        self.assertTrue(os.path.exists(rpath))

    def test_02_create(self):
        """Create and look at signals"""
        path = 'data/mailslot2'
        def catch_create_signal():
            return self.create_mailslot(path)
        (success, result) = self.signal_test(self.service_root,
            "MailSlotCreated", catch_create_signal)
        slot = MailSlot(path)
        self.assertTrue(success, 'Mailslot failed to Create')
        self.assertEqual(os.path.abspath(path), slot.path)
        self.assertTrue(result, 'Creation Signal Failed')
        self.assertEqual(str(result[0]), slot.slot_id, "Signal Returns SlotId")

    def test_03_config_loaded(self):
        """Loading a mailslot from a config"""
        path = 'data/mailslot0'
        slot = MailSlot(path)
        self.assertTrue(slot)

    def test_04_config_saved(self):
        """Saved new mailslot to config"""
        path = os.path.abspath('data/mailslot3')
        success = self.create_mailslot(path)
        import json
        done = False
        with open(CONF_FILE, 'r') as fhl:
            contents = json.loads(fhl.read())
            for slot_id, cpath in contents.get('slots', []):
                if cpath == path:
                    done = True
        self.assertTrue(done, "%s not in config" % path)

    def destruction(self):
        """Remove our mailslot"""
        for x in range(3):
            path = 'data/mailslot%d' % (x+1)
            if os.path.exists(path):
                if not self.service_root.DeleteMailSlot(path, True):
                    shutil.rmtree(path)



class AccessTestCase(DBusTestBase):
    """For accessing existing mailslots via dbus"""
    def creation(self):
        """Setup our mailslot object to test"""
        self.path = 'data/mailslot0'
        self.slot = MailSlot(self.path)

    def test_01_get_by_path(self):
        """Get a mailslot by it's path"""
        self.assertTrue(self.slot, 'No mailslot created to test')
        address = self.service_root.GetMailSlot(self.path)
        mailslot = dbus_object(address)
        self.assertTrue(mailslot)
        self.assertEqual(mailslot.GetId(), self.slot.slot_id)

    def test_03_get_by_id(self):
        """Get a mailslot by it's unique id"""
        self.assertTrue(self.slot, 'No mailslot created to test')
        one = dbus_object(self.service_root.GetMailSlot(self.path))
        two = dbus_object(self.service_root.GetMailSlot(one.GetId()))
        self.assertEqual(one.GetId(), two.GetId())
        
    def test_04_list_mailslots(self):
        """List all mailslots available"""
        self.assertTrue(self.slot, 'No mailslot created to test')
        slots = self.service_root.GetMailSlots()
        self.assertTrue(len(slots), 'No Slots Available')
        self.assertEqual(dbus_object(slots[0]).GetPath(), self.slot.path)

    def test_10_remove_mailslot(self):
        """Remove specified mailslot"""
        self.assertTrue(self.slot, 'No mailslot created to test')
        def catch_delete_signal():
            return self.service_root.DeleteMailSlot(self.path, False)
        (success, result) = self.signal_test(self.service_root,
            "MailSlotDeleted", catch_delete_signal)
        self.assertTrue(success, 'Deletion Failed')
        self.assertTrue(os.path.exists(self.path))
        self.assertEqual(str(self.slot.slot_id), str(result[0]))
        # Undo the damage
        success = self.service_root.CreateMailSlot(self.path)

    def test_11_purge_mailslot(self):
        """Purge specified mailslot"""
        self.assertTrue(self.slot, 'No mailslot created to test')
        def catch_delete_signal():
            return self.service_root.DeleteMailSlot(self.path, True)
        (success, result) = self.signal_test(self.service_root,
            "MailSlotDeleted", catch_delete_signal)
        self.assertTrue(success, 'Deletion Failed')
        self.assertFalse(os.path.exists(self.path))
        self.assertEqual(str(self.slot.slot_id), str(result[0]))


if __name__ == '__main__':
    prepare_data([CONF_FILE, 'data/mailslot0'], True)
    Pipette.run_unittest( CreationTestCase, AccessTestCase )

